/**********************************************************************

  Audacity: A Digital Audio Editor

  Export.cpp

  Dominic Mazzoni

**********************************************************************/

#include <wx/button.h>
#include <wx/textctrl.h>
#include <wx/choice.h>
#include <wx/file.h>
#include <wx/thread.h>
#include <wx/radiobut.h>
#include <wx/sizer.h>
#include <wx/timer.h>
#include <wx/filedlg.h>
#include <wx/msgdlg.h>
#include <wx/progdlg.h>
#include <wx/textfile.h>

#include "snd/snd.h"

#include "AudacityApp.h"
#include "WaveTrack.h"
#include "Track.h"
#include "DirManager.h"
#include "Prefs.h"
#include "Mix.h"

#include "Export.h"

bool Export(TrackList *tracks, bool selectionOnly,
	    double t0, double t1)
{
  wxString format =
    gPrefs->Read("/FileFormats/DefaultExportFormat", "WAV");    

  int header = SND_HEAD_NONE;

  if (format=="WAV")
    header = SND_HEAD_WAVE;
  else if (format=="AIFF")
    header = SND_HEAD_AIFF;
  else if (format=="IRCAM")
    header = SND_HEAD_IRCAM;
  else if (format=="AU")
    header = SND_HEAD_NEXT;

  if (header == SND_HEAD_NONE) {
    wxMessageBox(wxString::Format("Sorry, you cannot export %s files (yet).\n"
				  "Open the Preferences and change your "
				  "default export format.",
				  (const char *)format));
    return false;
  }

  TrackListIterator iter1(tracks);
  int numSelected = 0, numLeft = 0, numRight = 0, numMono = 0;
  double rate = 0;
  
  VTrack *tr = iter1.First();
  while(tr) {
    if (tr->GetKind() == VTrack::Wave) {
      if (tr->selected || !selectionOnly) {

	if (rate == 0)
	  rate = ((WaveTrack *)tr)->GetRate();
	if (rate != ((WaveTrack *)tr)->GetRate()) {
	  wxMessageBox("Cannot export tracks with different rates.");
	  return false;
	}

	numSelected++;
	if (tr->channel == VTrack::LeftChannel)
	  numLeft++;
	if (tr->channel == VTrack::RightChannel)
	  numRight++;
	if (tr->channel == VTrack::MonoChannel)
	  numMono++;
      }
    }
    tr = iter1.Next();
  }

  if (numSelected == 0 && selectionOnly) {
    wxMessageBox("No tracks are selected!\n"
		 "Choose Export... to export all tracks.");
    return false;
  }

  bool stereo = false;
  if (numRight > 0 || numLeft > 0)
    stereo = true;

  numRight += numMono;
  numLeft += numMono;
  
  if (numLeft > 1 || numRight > 1)
    if (stereo)
      wxMessageBox("Your tracks will be mixed down to two stereo channels "
		   "in the exported file.");
    else
      wxMessageBox("Your tracks will be mixed down to a single mono channel "
		   "in the exported file.");    
  
  wxString extension = "." + format;
  
  wxString fName =
    wxFileSelector(wxString::Format("Save %s File As:",
				    (const char *)format),
		   NULL,
		   extension,  // default file name
		   extension,  // extension
		   "*.*",
		   wxSAVE);
  
  if (fName.Length() >= 256) {
    wxMessageBox("Sorry, pathnames longer than 256 characters not supported.");
    return false;
  }
  
  if (fName == "")
    return false;

  // Use snd library to export file

  snd_node sndfile;
  snd_node sndbuffer;

  sndfile.device = SND_DEVICE_FILE;
  sndfile.write_flag = SND_WRITE;
  strcpy(sndfile.u.file.filename, (const char *)fName);
  sndfile.u.file.file = 0;
  sndfile.u.file.header = header;
  sndfile.u.file.byte_offset = 0;
  sndfile.u.file.end_offset = 0;
  sndfile.u.file.swap = 0;
  sndfile.format.channels = stereo? 2: 1;
  sndfile.format.mode = SND_MODE_PCM; // SND_MODE_FLOAT
  sndfile.format.bits = 16;
  sndfile.format.srate = int(rate+0.5);
  
  int err;
  long flags=0;

  err = snd_open(&sndfile, &flags);
  if (err) {
	wxMessageBox("Could not write to file.");
	return false;
  }
  
  sndbuffer.device = SND_DEVICE_MEM;
  sndbuffer.write_flag = SND_READ;
  sndbuffer.u.mem.buffer_max = 0;
  sndbuffer.u.mem.buffer = 0;
  sndbuffer.u.mem.buffer_len = 0;
  sndbuffer.u.mem.buffer_pos = 0;
  sndbuffer.format.channels = stereo? 2: 1;
  sndbuffer.format.mode = SND_MODE_PCM; // SND_MODE_FLOAT
  sndbuffer.format.bits = 16;
  sndbuffer.format.srate = int(rate+0.5);

  double timeStep = 10.0; // write in blocks of 10 secs
  
  sampleCount maxSamples = int(timeStep * rate + 0.5);

  wxProgressDialog *progress = NULL;  
  wxYield();
  wxStartTimer();
  wxBusyCursor busy;

  double t = t0;

  while(t < t1) {

    double deltat = timeStep;
    if (t + deltat > t1)
      deltat = t1 - t;

    sampleCount numSamples = int(deltat * rate + 0.5);

    Mixer *mixer = new Mixer(stereo? 2: 1, numSamples, true);
    wxASSERT(mixer);
    mixer->Clear();

    char *buffer = new char[numSamples*2*sndbuffer.format.channels];
    wxASSERT(buffer);

    TrackListIterator iter2(tracks);
    tr = iter2.First();
    while(tr) {
      if (tr->GetKind() == VTrack::Wave) {
	if (tr->selected || !selectionOnly) {
	  if (tr->channel == VTrack::MonoChannel)
	    mixer->MixMono((WaveTrack *)tr, t, t+deltat);
	  if (tr->channel == VTrack::LeftChannel)
	    mixer->MixLeft((WaveTrack *)tr, t, t+deltat);
	  if (tr->channel == VTrack::RightChannel)
	    mixer->MixRight((WaveTrack *)tr, t, t+deltat);
	}
      }
      tr = iter2.Next();
    }

    sampleType *mixed = mixer->GetBuffer();
    
    long b2 = snd_convert(&sndfile, buffer,  // to
			  &sndbuffer, mixed, numSamples); // from

    int actual = snd_write(&sndfile, buffer, b2);

    t += deltat;

    if (!progress && wxGetElapsedTime(false) > 500) {
      progress =
	new wxProgressDialog("Export",
			     wxString::Format("Exporting %s file",
					      (const char *)format),
			     1000);
    }	
    if (progress) {
      progress->Update(int(((t-t0)*1000)/(t1-t0)+0.5));
    }
	
    delete mixer;
    delete[] buffer;
  }

  snd_close(&sndfile);
  
#ifdef __WXMAC__
  FSSpec spec ;

  wxUnixFilename2FSSpec( fName , &spec ) ;
  FInfo finfo ;
  if ( FSpGetFInfo( &spec , &finfo ) == noErr )
  {
    switch(header) {
    case SND_HEAD_AIFF:
  	  finfo.fdType = 'AIFF';
  	  break;
    case SND_HEAD_IRCAM:
  	  finfo.fdType = 'IRCA';
  	  break;
    case SND_HEAD_NEXT:
  	  finfo.fdType = 'AU  ';
  	  break;
    case SND_HEAD_WAVE:
  	  finfo.fdType = 'WAVE';
  	  break;
    }
  	
    finfo.fdCreator = AUDACITY_CREATOR;

    FSpSetFInfo( &spec , &finfo ) ;
  }
#endif

  if (progress)
    delete progress;
  
  return true;
}

/*
  if (header == SND_HEAD_WAVE+1) {
	// Export as text

	wxTextFile f(fName);
	#ifdef __WXMAC__
    wxFile *temp = new wxFile();
    temp->Create(fName);
    delete temp;
    #else
    f.Create();
    #endif
	f.Open();
	if (!f.IsOpened()) {
	  wxMessageBox("Couldn't write to "+fName);
	  return false;
	}

	int len = left->numSamples;

	if (len > 1000000) {
	  wxMessageBox("Sorry, can't export text files more than 1MB");
	  return false;
	}

	sampleType *buffer = new sampleType[len];
	left->Get(buffer, 0, len);

	for(int i=0; i<len; i++)
	  f.AddLine(wxString::Format("%g", buffer[i]/32767.0));

	f.Write();
	f.Close();

	return true;
  }
*/
